# mypackage.__init__ and mypackage.a are collected into PYZ archive (the default behavior), while mypackage.b is
# collected as source .py file only.
module_collection_mode = {
    'mypackage.b': 'py',
}
